import os, time, shutil
import torch
from transformers import DynamicCache
from typing import Callable, Optional, Tuple, Union, Dict, Any, Iterable, List

class oCache:	
	def ini_ocache(self, cache_dir, stats):
		if not cache_dir: raise Error("cache_dir can not be empty. If you are trying to not use DiskCache, simply set past_key_values=None. This will use default DynamicCache")
		self.cache_folder = os.path.join(cache_dir, "kv_cache")
		self.key_cache2, self.value_cache2 = [], []
		if os.path.exists(self.cache_folder): shutil.rmtree(self.cache_folder)
		os.makedirs(self.cache_folder)
		self.stats = stats

	def load_from_disk(self, layer_idx, device="cuda:0"):
		path = f"{self.cache_folder}/layer_{layer_idx}.pt"
		if not os.path.exists(path): return None
		t1 = time.perf_counter()
		tensors = torch.load(path, map_location=device)
		if self.stats: self.stats.set("kvload", t1)
		return tensors

	def save_to_disk(self, tensors, layer_idx):
		t1 = time.perf_counter()
		path = f"{self.cache_folder}/layer_{layer_idx}.pt"
		tensors = (tensors[0].cpu(), tensors[1].cpu())
		torch.save(tensors, path)
		if self.stats: self.stats.set("kvsave", t1)


class KVCache(DynamicCache, oCache): #DiskCache
	def __init__(self, cache_dir="./kv_cache", stats=None):
		super().__init__()		
		self.ini_ocache(cache_dir, stats)

	def update(
		self,
		key_states: torch.Tensor,
		value_states: torch.Tensor,
		layer_idx: int,
		cache_kwargs: Optional[Dict[str, Any]] = None,
	) -> Tuple[torch.Tensor, torch.Tensor]:
		tensors = self.load_from_disk(layer_idx)
		if tensors is not None:
			self.layers[layer_idx].keys, self.layers[layer_idx].values = tensors
			if layer_idx < len(self.key_cache2):
				self.layers[layer_idx].keys = torch.cat([self.layers[layer_idx].keys, self.key_cache2[layer_idx]], dim=-2)
				self.layers[layer_idx].values = torch.cat([self.layers[layer_idx].values, self.value_cache2[layer_idx]], dim=-2)
				self.key_cache2[layer_idx] = torch.cat([self.key_cache2[layer_idx], key_states], dim=-2)
				self.value_cache2[layer_idx] = torch.cat([self.value_cache2[layer_idx], value_states], dim=-2)				
			else:
				self.key_cache2.append(key_states)
				self.value_cache2.append(value_states)
		
		out = super().update(key_states, value_states, layer_idx, cache_kwargs) #tuple of (self.key_cache[layer_idx], self.value_cache[layer_idx])		
		if tensors is None: self.save_to_disk(out, layer_idx) #save only first time cause it's slow to save
		self.layers[layer_idx].keys, self.layers[layer_idx].values = None, None
		return out



class KVCache_legacy(DynamicCache):
	def update(
		self,
		key_states: torch.Tensor,
		value_states: torch.Tensor,
		layer_idx: int,
		cache_kwargs: Optional[Dict[str, Any]] = None,
	) -> Tuple[torch.Tensor, torch.Tensor]:
		tensors = self.load_from_disk(layer_idx)
		if tensors is not None:
			self.key_cache[layer_idx], self.value_cache[layer_idx] = tensors
			if layer_idx < len(self.key_cache2):
				self.key_cache[layer_idx] = torch.cat([self.key_cache[layer_idx], self.key_cache2[layer_idx]], dim=-2)
				self.value_cache[layer_idx] = torch.cat([self.value_cache[layer_idx], self.value_cache2[layer_idx]], dim=-2)
				self.key_cache2[layer_idx] = torch.cat([self.key_cache2[layer_idx], key_states], dim=-2)
				self.value_cache2[layer_idx] = torch.cat([self.value_cache2[layer_idx], value_states], dim=-2)				
			else:
				self.key_cache2.append(key_states)
				self.value_cache2.append(value_states)
		
		out = super().update(key_states, value_states, layer_idx, cache_kwargs) #tuple of (self.key_cache[layer_idx], self.value_cache[layer_idx])		
		if tensors is None: self.save_to_disk(out, layer_idx) #save only first time cause it's slow to save
		self.key_cache[layer_idx], self.value_cache[layer_idx] = torch.empty(0), torch.empty(0)
		return out



import json, os, time, math, re
import torch
from torch.utils.dlpack import from_dlpack
import cupy as cp
import kvikio
#from safetensors.torch import safe_open, load_file
import struct

stats = None

DTYPE_MAP = {
	"float16": cp.float16,
	"bfloat16": cp.float16, #cp.dtype('bfloat16'),
	"float32": cp.float32,
	"float64": cp.float64,
	"int8": cp.int8,
	"int32": cp.int32,
}

class GDSWeights:
	def __init__(self, path: str, device="cuda:0"):
		self.path = path #<model_dir>/gds_export/
		manifest_path = os.path.join(path, 'manifest.json')
		with open(manifest_path) as f:
			self.manifest = json.load(f)
		self.device = torch.device(device)
		self.offloaded_map = {}

	def load_param_to_cuda(self, name: str) -> torch.Tensor:
		meta = self.manifest[name]
		path, shape, dtype = os.path.join(self.path, meta["path"]), meta["shape"], meta["dtype"]
		t = self.get_offloaded_from_cpu_to_cuda(name)
		if t is not None: return t

		if meta.get("packed")=="mxfp4":
			return self.load_mxfp4_from_disk(path, shape, dtype)
		elif meta.get("dtype").startswith("torch"):
			return self.load_torch_from_disk(path)
		else: #kvikio, numpy
			return self.load_from_disk_to_cuda(path, shape, dtype)

	def load_from_disk_to_cuda(self, path, shape, dtype): #str, list, str
		cp_dtype = DTYPE_MAP[dtype]
		n_elems = 1
		for s in shape:
			n_elems *= s
		nbytes = n_elems * cp.dtype(cp_dtype).itemsize

		# Allocate on GPU
		with cp.cuda.Device(0):
			buf = cp.empty(n_elems, dtype=cp_dtype)

		# DMA read directly into GPU buffer
		with kvikio.CuFile(path, "r") as f:
			# Read raw bytes straight into GPU memory
			n = f.read(buf)
			if n != nbytes:
				raise IOError(f"Short read: {n} of {nbytes} bytes from {path}")

		# Reshape and hand to torch via DLPack
		buf = buf.reshape(shape)
		t = from_dlpack(buf.toDlpack())  # torch.cuda.Tensor shares memory
		return t    

	def has(self, name: str) -> bool:
		return name in self.manifest

	def load_torch_from_disk(self, path):
		tensor = torch.load(path, map_location=self.device)
		return tensor

	def load_mxfp4_from_disk(self, path, shape, dtype):
		packed = torch.load(path, map_location=self.device) #{_blocks:t, _scales:t}		
		tensor = convert_moe_packed_tensors(packed["_blocks"], packed["_scales"]).to(self.device)		
		return tensor

	def offload_param_to_cpu(self, name):
		meta = self.manifest[name]
		path, shape, dtype, packed = os.path.join(self.path, meta["path"]), meta["shape"], meta["dtype"], meta.get("packed")
		if packed=="mxfp4" or dtype.startswith("torch"):
			tensor = torch.load(path, map_location="cpu")
		else: #kvikio, numpy
			tensor = self.load_from_disk_to_cuda(path, shape, dtype).cpu() #should be without GPU
		self.offloaded_map[name] = {"shape":shape, "dtype":dtype, "packed":packed, "tensor":tensor}

	def get_offloaded_from_cpu_to_cuda(self, name):
		if name in self.offloaded_map:
			meta = self.offloaded_map[name]
			t, packed = meta["tensor"], meta["packed"]
			t1 = time.perf_counter()
			if packed=="mxfp4":
				tensor = convert_moe_packed_tensors(t["_blocks"].to(self.device), t["_scales"].to(self.device))
			else:
				tensor = t.to(self.device)
			if stats: stats.set("offloaded_cpu_to_cuda", t1)
			return tensor
		return None

#=========================================================================

FP4_VALUES = [
	+0.0,
	+0.5,
	+1.0,
	+1.5,
	+2.0,
	+3.0,
	+4.0,
	+6.0,
	-0.0,
	-0.5,
	-1.0,
	-1.5,
	-2.0,
	-3.0,
	-4.0,
	-6.0,
]

def convert_moe_packed_tensors( #copied from transformers/integrations/mxfp4.py
	blocks,
	scales,
	*,
	dtype: torch.dtype = torch.bfloat16,
	rows_per_chunk: int = 32768 * 1024,  # TODO these values are not here by mistake ;)
) -> torch.Tensor:
	"""
	Convert the mxfp4 weights again, dequantizing and makes them compatible with the forward
	pass of GPT_OSS.
	"""

	# Check if blocks and scales are on CPU, and move to GPU if so
	#if not blocks.is_cuda and torch.cuda.is_available():
	#    blocks = blocks.cuda()
	#    scales = scales.cuda()

	scales = scales.to(torch.int32) - 127  # TODO that's because 128=2**7

	assert blocks.shape[:-1] == scales.shape, f"{blocks.shape[:-1]=} does not match {scales.shape=}"

	lut = torch.tensor(FP4_VALUES, dtype=dtype, device=blocks.device)

	*prefix_shape, G, B = blocks.shape
	rows_total = math.prod(prefix_shape) * G

	blocks = blocks.reshape(rows_total, B)
	scales = scales.reshape(rows_total, 1)

	out = torch.empty(rows_total, B * 2, dtype=dtype, device=blocks.device)

	for r0 in range(0, rows_total, rows_per_chunk):
		r1 = min(r0 + rows_per_chunk, rows_total)

		blk = blocks[r0:r1]
		exp = scales[r0:r1]

		# nibble indices -> int64
		idx_lo = (blk & 0x0F).to(torch.long)
		idx_hi = (blk >> 4).to(torch.long)

		sub = out[r0:r1]
		sub[:, 0::2] = lut[idx_lo]
		sub[:, 1::2] = lut[idx_hi]

		torch.ldexp(sub, exp, out=sub)
		del idx_lo, idx_hi, blk, exp, sub

	out = out.reshape(*prefix_shape, G, B * 2).view(*prefix_shape, G * B * 2)
	del blocks, scales, lut
	return out.transpose(1, 2).contiguous()


#=========================================================================

class asyncCudaLoader: #it's slow, maybe need to adapt for experts common load
	def async_move_to_gpu(self, tensor, stream=None):		
		if not tensor.is_pinned():
			try:
				tensor = tensor.pin_memory()
			except Exception:
				# some dtypes or storages may not support pinning; fall back gracefully
				pass

		if stream is None:
			return tensor.to(device=self.device, non_blocking=True)

		with torch.cuda.stream(stream):
			dst = tensor.to(device=self.device, non_blocking=True)
		return dst

	def load_pt_to_gpu(self, pt_path, max_workers=4):
		gpu_state = {}
		obj = torch.load(pt_path, map_location="cpu")		
		streams = [torch.cuda.Stream(device=self.device) for _ in range(max_workers)]
		executor = ThreadPoolExecutor(max_workers=max_workers)
		
		def handle_item(key, val, stream_idx):
			stream = streams[stream_idx % len(streams)]
			return self.async_move_to_gpu(val, stream=stream)

		futures, idx = {}, 0
		for k, v in obj.items():
			futures[k] = executor.submit(handle_item, k, v, idx)
			idx += 1

		# collect results
		for k, fut in futures.items(): gpu_state[k] = fut.result()

		# synchronize all streams to ensure transfers are finished
		for s in streams: s.synchronize()
		executor.shutdown(wait=True)
		return gpu_state


class MoEWeightsLoader(GDSWeights): #qwen3_next base.pt(dict: attr=>tensor)
	def load_dict_to_cuda(self, base):
		t = self.get_offloaded_dict_to_cuda(base)
		if t: return t
		return self.load_dict_from_disk(base, device=self.device)

	def offload_dict_to_gpu_cpu(self, base, gpu=False):
		d = self.load_dict_from_disk(base, device=self.device if gpu else 'cpu')
		self.offloaded_map[base] = d

	def get_offloaded_dict_to_cuda(self, base):
		if base in self.offloaded_map:
			d, d2 = self.offloaded_map[base], {}
			for attr_path, tensor in d.items():
				d2[attr_path] = tensor.to(self.device, non_blocking=True)
			return d2
		return None
	
	def load_dict_from_disk(self, base, device='cpu'):
		return torch.load(self.path+base.replace(".","__")+".pt", map_location=device)

#=========================================================================

class SafeTensorReader: #safetensors replacement because its mmap is killing the RAM
	def __init__(self, path):
		self.path = path		
		with open(path, "rb") as f:
			header_len = struct.unpack("<Q", f.read(8))[0]
			self.header = json.loads(f.read(header_len))
			self.data_offset = 8 + header_len
		self._fp = open(path, "rb")
		self.DTYPE_MAP = {"F32": torch.float32, "F16": torch.float16, "BF16": torch.bfloat16}
	
	def close(self):
		self._fp.close()

	def keys(self):
		return list(self.header.keys())

	def get_tensor(self, name):
		info = self.header[name]
		dtype = self.DTYPE_MAP[info["dtype"]]
		shape = info["shape"]
		off0, off1 = info["data_offsets"]
		self._fp.seek(self.data_offset + off0)
		buf = self._fp.read(off1 - off0)
		return torch.frombuffer(memoryview(buf), dtype=dtype).reshape(shape)


class MoEWeightsLoader2(MoEWeightsLoader): #qwen3_next safetensors
	def __init__(self, path: str, device="cuda:0"):
		self.path = path #<model_dir>
		index_path = os.path.join(path, 'model.safetensors.index.json')
		with open(index_path) as f: indexes = json.load(f)
		self.manifest, self.safetensors = {}, {}
		for manifest_name, filename in indexes["weight_map"].items():
			match1 = re.search(r"(model\.layers\.\d+\.mlp\.experts\.\d+\.)", manifest_name)
			match2 = re.search(r"(model\.layers\.\d+\.)", manifest_name)
			if match1 or match2:
				base = match1.group(1) if match1 else match2.group(1)
				if base not in self.manifest: self.manifest[base] = {}
				attr_path = manifest_name.replace(base, "")
				self.manifest[base][attr_path] = filename

		self.device = torch.device(device)
		self.offloaded_map = {}

	def load_dict_from_disk(self, base, device='cpu'): #original safetensors
		dbase, d = self.manifest[base], {}
		for attr_path, filename in dbase.items():
			d[attr_path] = self.safetensors[filename].get_tensor(base+attr_path).to(device)
		return d

	def preload_layer_safetensors(self, base): #load only couple instead of 48
		#for filename, x in self.safetensors.items(): x.close() #f.__exit__(None, None, None)
		#del self.safetensors
		#self.safetensors = {}
		for base1 in list(self.manifest.keys()):
			if base1.startswith(base):
				for attr_path, filename in self.manifest[base1].items():
					if filename not in self.safetensors:
						filepath = os.path.join(self.path, filename)
						self.safetensors[filename] = SafeTensorReader(filepath) #safe_open(filepath, framework="pt")

#=========================================================================

if __name__=="__main__":
	q = GDSWeights("/media/mega4alik/ssd/models/gpt-oss-20B/gds_export/")
	t = q.load_param_to_cuda("model.layers.0.self_attn.q_proj.weight")
	print(t, t.dtype, t.shape)



import os, requests, zipfile
import torch
from transformers import AutoTokenizer
from .utils import Stats, file_get_contents
from .gds_loader import GDSWeights, MoEWeightsLoader2
from .kvcache import KVCache

class Inference:
	def __init__(self, model_id, device="cuda:0", logging=True):
		self.model_id = model_id
		self.device = torch.device(device)
		self.stats = Stats() if logging else None

	def download_and_unpack(self, models_dir: str):
		os.makedirs(models_dir, exist_ok=True)
		urls = {
			"llama3-1B-chat": "https://ollm.s3.us-east-1.amazonaws.com/models/llama3-1B-chat.zip",
			"llama3-3B-chat": "https://ollm.s3.us-east-1.amazonaws.com/models/llama3-3B-chat.zip",
			"llama3-8B-chat": "https://ollm.s3.us-east-1.amazonaws.com/models/llama3-8B-chat.zip",
			"gpt-oss-20B":    "https://ollm.s3.us-east-1.amazonaws.com/models/gpt-oss-20B.zip"
		}
		url = urls[self.model_id]
		
		# Extract filename from URL
		filename = url.split("/")[-1]
		zip_path = os.path.join(models_dir, filename)

		# Download the file
		print(f"Downloading {url} ...")
		response = requests.get(url, stream=True)
		response.raise_for_status()
		with open(zip_path, "wb") as f:
			for chunk in response.iter_content(chunk_size=8192):
				f.write(chunk)
		print(f"Downloaded to {zip_path}")

		# Unzip
		print(f"Unpacking {zip_path} ...")
		with zipfile.ZipFile(zip_path, 'r') as zip_ref:
			zip_ref.extractall(models_dir)
		print(f"Unpacked to {models_dir}")

		os.remove(zip_path) # Optional: remove the zip file after extraction

	
	def hf_download(self, model_dir):
		from huggingface_hub import snapshot_download
		urls = {"qwen3-next-80B": "Qwen/Qwen3-Next-80B-A3B-Instruct"}
		url = urls[self.model_id]
		print(f"Downloading {url} ...")
		snapshot_download(
		    repo_id=url,
		    local_dir=model_dir,
		    local_dir_use_symlinks=False
		)

	
	def ini_model(self, models_dir="./models/", force_download=False):
		models_list = ["llama3-1B-chat", "llama3-3B-chat", "llama3-8B-chat", "gpt-oss-20B", "qwen3-next-80B"]
		if self.model_id not in models_list:
			raise ValueError("Incorrect model id. It must be one of", models_list)
		
		model_dir = os.path.join(models_dir, self.model_id)
		if os.path.exists(model_dir)==False or force_download==True:
			if self.model_id=="qwen3-next-80B": 
				self.hf_download(model_dir)
			else:
				self.download_and_unpack(models_dir)
		
		print("loading model from", model_dir)
		if self.model_id=="qwen3-next-80B":
			from . import qwen3_next
			qwen3_next.loader = MoEWeightsLoader2(model_dir)
			qwen3_next.stats = self.stats
			self.model = qwen3_next.MyQwen3NextForCausalLM.from_pretrained(model_dir, torch_dtype=torch.bfloat16, device_map="cpu", attn_implementation="flash_attention_2", low_cpu_mem_usage=True, ignore_mismatched_sizes=True)
		elif self.model_id=="gpt-oss-20B":
			from . import gpt_oss
			gpt_oss.loader = GDSWeights(os.path.join(model_dir, "gds_export"))
			gpt_oss.stats = self.stats
			self.model = gpt_oss.MyGptOssForCausalLM.from_pretrained(model_dir, torch_dtype=torch.bfloat16, device_map="cpu", low_cpu_mem_usage=True, ignore_mismatched_sizes=True)
		else:
			from . import llama
			llama.loader = GDSWeights(os.path.join(model_dir, "gds_export"))
			llama.stats = self.stats			
			self.model = llama.MyLlamaForCausalLM.from_pretrained(model_dir, torch_dtype=torch.float16, device_map="cpu", attn_implementation="flash_attention_2", low_cpu_mem_usage=True, ignore_mismatched_sizes=True)
			self.model.clean_layers_weights()

		self.model.eval()
		self.model.to(self.device)
		self.tokenizer = AutoTokenizer.from_pretrained(model_dir)

	
	def offload_layers_to_cpu(self, **args):
		self.model.offload_layers_to_cpu(**args)
	
	def offload_layers_to_gpu_cpu(self, **args):
		self.model.offload_layers_to_gpu_cpu(**args)
	
	def DiskCache(self, cache_dir="./kvcache"):
		if self.model_id in ["gpt-oss-20B"]:
			print(f"{self.model_id} DiskCache is not supported at the moment. Using default DynamicCache instead")
			return None
		elif self.model_id=="qwen3-next-80B":
			from .qwen3_next import Qwen3NextDiskCache
			return Qwen3NextDiskCache(self.model.config, cache_dir=cache_dir, stats=self.stats)
		else:
			return KVCache(cache_dir=cache_dir, stats=self.stats) #config=?



# efficiant Llama that runs on consumer PC with 8GB VRAM

import time, os
from datetime import datetime
import threading
import numpy as np
import torch
from torch import nn
from typing import Callable, Optional, Tuple, Union, Dict, Any, Iterable, List, Unpack
from transformers import LlamaForCausalLM, Cache

from .utils import _walk_to_parent, _assign_tensor_to_module, _set_meta_placeholder
from .gds_loader import GDSWeights
from .attention import online_chunked_grouped_attention_rope_no_mask as chunked_attention

#global vars
loader, stats = None, None

#======== rewriting core classes (tested on transformers==4.52.3) ==============
from transformers.models.llama.modeling_llama import apply_rotary_pos_emb, eager_attention_forward, LlamaAttention, LlamaMLP, LlamaDecoderLayer, LlamaModel, LlamaConfig, create_causal_mask
from transformers.modeling_outputs import BaseModelOutputWithPast

class MyLlamaAttention(LlamaAttention):
	def forward(
		self,
		hidden_states: torch.Tensor,
		position_embeddings: Tuple[torch.Tensor, torch.Tensor],
		attention_mask: Optional[torch.Tensor],
		past_key_value: Optional = None, #[Cache]
		cache_position: Optional[torch.LongTensor] = None,
		**kwargs: Optional,  #Unpack[FlashAttentionKwargs], #kwargs={'position_ids': tensor([[44]], device='cuda:0'), 'output_attentions': False, 'use_cache': True}
	) -> Tuple[torch.Tensor, Optional[torch.Tensor], Optional[Tuple[torch.Tensor]]]:
		input_shape = hidden_states.shape[:-1]
		hidden_shape = (*input_shape, -1, self.head_dim)

		query_states = self.q_proj(hidden_states).view(hidden_shape).transpose(1, 2)
		key_states = self.k_proj(hidden_states).view(hidden_shape).transpose(1, 2)
		value_states = self.v_proj(hidden_states).view(hidden_shape).transpose(1, 2)		

		cos, sin = position_embeddings		
		query_states, key_states = apply_rotary_pos_emb(query_states, key_states, cos, sin)

		if past_key_value is not None:
			# sin and cos are specific to RoPE models; cache_position needed for the static cache
			cache_kwargs = {"sin": sin, "cos": cos, "cache_position": cache_position}
			key_states, value_states = past_key_value.update(key_states, value_states, self.layer_idx, cache_kwargs)

		#===		
		attn_output = chunked_attention(query_states, key_states, value_states, position_ids=kwargs["position_ids"], q_block_size=32768, k_block_size=(1024 if input_shape[1] > 128 else 1000000)).transpose(1, 2)
		attn_weights = None
		"""
		attention_interface: Callable = eager_attention_forward
		attn_output, attn_weights = attention_interface(
			self,
			query_states,
			key_states,
			value_states,
			attention_mask,
			dropout=0.0 if not self.training else self.attention_dropout,
			scaling=self.scaling,
			**kwargs,
		)
		"""
		#print(attn_output1.shape, attn_output.shape)
		#print("Error:", (attn_output1 - attn_output).abs().max().item())
		del query_states, key_states, value_states
		#===
		attn_output = attn_output.reshape(*input_shape, -1).contiguous()
		attn_output = self.o_proj(attn_output)
		return attn_output, attn_weights


class MyLlamaMLP(LlamaMLP):
	def forward(self, x): #down_proj = self.down_proj(self.act_fn(self.gate_proj(x)) * self.up_proj(x))
		chunk_size, chunks = 16384, []
		x = x.squeeze(0)
		for i in range(0, x.shape[0], chunk_size):
			gate_chunk = self.act_fn(self.gate_proj(x[i:i+chunk_size]))			
			up_chunk = self.up_proj(x[i:i+chunk_size])
			out_chunk = self.down_proj(gate_chunk * up_chunk)
			chunks.append(out_chunk)
		down_proj = torch.cat(chunks, dim=0).unsqueeze(0) #T,C->1,T,C
		return down_proj


class MyLlamaDecoderLayer(LlamaDecoderLayer):
	def __init__(self, config: LlamaConfig, layer_idx: int):
		self.layer_idx = layer_idx
		super().__init__(config, layer_idx)

	def _layer_param_manifest_names(self) -> dict:
		"""
		Return a mapping of local attribute paths -> manifest names in your storage.
		Adjust the keys/names for your model's param naming convention.
		Example keys are relative to `self` (not full HF names).
		"""
		base = f"model.layers.{self.layer_idx}"
		return {
			"self_attn.q_proj.weight": f"{base}.self_attn.q_proj.weight",
			"self_attn.k_proj.weight": f"{base}.self_attn.k_proj.weight",
			"self_attn.v_proj.weight": f"{base}.self_attn.v_proj.weight",
			"self_attn.o_proj.weight": f"{base}.self_attn.o_proj.weight",
			"mlp.gate_proj.weight":  f"{base}.mlp.gate_proj.weight",
			"mlp.up_proj.weight":    f"{base}.mlp.up_proj.weight",
			"mlp.down_proj.weight":  f"{base}.mlp.down_proj.weight",
			"input_layernorm.weight":  f"{base}.input_layernorm.weight",
			"post_attention_layernorm.weight": f"{base}.post_attention_layernorm.weight",
			# add any biases or additional params you need
		}    

	def _load_layer_weights(self):
		manifest_map = self._layer_param_manifest_names()
		for attr_path, manifest_name in manifest_map.items():
			try:
				t1 = time.perf_counter()
				tensor = loader.load_param_to_cuda(manifest_name)
				parent, leaf = _walk_to_parent(self, attr_path)
				_assign_tensor_to_module(parent, leaf, tensor)
				if stats: stats.set("layer_load", t1)
			except Exception as e:
				# Be explicit about failures so you can debug missing names
				raise RuntimeError(f"failed to load {manifest_name} into {attr_path}: {e}")
		#if torch.cuda.is_available(): torch.cuda.synchronize()

	def _unload_layer_weights(self):
		"""Replace each loaded attribute with a meta placeholder to free GPU memory."""
		manifest_map = self._layer_param_manifest_names()
		for attr_path in manifest_map.keys():
			parent, leaf = _walk_to_parent(self, attr_path)
			# replace with placeholder (keeps module graph intact)
			_set_meta_placeholder(parent, leaf)

	def forward(self, *args, **kwargs):
		self._load_layer_weights()
		out = super().forward(*args, **kwargs)
		self._unload_layer_weights()
		return out



class MyLlamaModel(LlamaModel):
	#self.parent_lm_head is set

	def forward(
		self,
		input_ids: Optional[torch.LongTensor] = None,
		attention_mask: Optional[torch.Tensor] = None,
		position_ids: Optional[torch.LongTensor] = None,
		past_key_values: Optional[Cache] = None,
		inputs_embeds: Optional[torch.FloatTensor] = None,
		cache_position: Optional[torch.LongTensor] = None,
		use_cache: Optional[bool] = None,
		**kwargs: Unpack, #[TransformersKwargs]
	) -> BaseModelOutputWithPast:
		if (input_ids is None) ^ (inputs_embeds is not None):
			raise ValueError("You must specify exactly one of input_ids or inputs_embeds")

		if inputs_embeds is None:
			inputs_embeds: torch.Tensor = self.embed_tokens(input_ids)

		if use_cache and past_key_values is None:
			past_key_values = DynamicCache()

		if cache_position is None:
			past_seen_tokens = past_key_values.get_seq_length() if past_key_values is not None else 0
			cache_position: torch.Tensor = torch.arange(
				past_seen_tokens, past_seen_tokens + inputs_embeds.shape[1], device=inputs_embeds.device
			)

		if position_ids is None:
			position_ids = cache_position.unsqueeze(0)

		causal_mask = create_causal_mask(
			config=self.config,
			input_embeds=inputs_embeds,
			attention_mask=attention_mask,
			cache_position=cache_position,
			past_key_values=past_key_values,
			position_ids=position_ids,
		)

		hidden_states = inputs_embeds
		position_embeddings = self.rotary_emb(hidden_states, position_ids)

		#============= meine ==============		
		self.embed_tokens.cpu(); self.parent_lm_head.cpu()		

		for layer_idx, decoder_layer in enumerate(self.layers[: self.config.num_hidden_layers]):						
			p1 = None # 1NextInThread
			"""
			if layer_idx+1 < len(self.layers):
				p1 = threading.Thread(target=self.layers[layer_idx+1]._load_layer_weights, args=())
				p1.start()
			if layer_idx==0: decoder_layer._load_layer_weights()			
			"""
			hidden_states = decoder_layer(
				hidden_states,
				attention_mask=causal_mask,
				position_ids=position_ids,
				past_key_value=past_key_values,
				cache_position=cache_position,
				position_embeddings=position_embeddings,
				**kwargs,
			)
			if p1 is not None: p1.join()

		hidden_states = self.norm(hidden_states)
		self.embed_tokens.to(hidden_states.device); self.parent_lm_head.to(hidden_states.device)
		if stats: print("./Llama.forward.", datetime.now().strftime("%H:%M:%S"), stats.print_and_clean() if stats else "")
		#====================================
		
		return BaseModelOutputWithPast(
			last_hidden_state=hidden_states,
			past_key_values=past_key_values,
		)

# Monkey-patch
import transformers.models.llama.modeling_llama as llama_modeling
#llama_modeling.LlamaAttention = MyLlamaAttention #replaced to stable attn_implementation="flash_attention_2"
llama_modeling.LlamaMLP = MyLlamaMLP
llama_modeling.LlamaDecoderLayer = MyLlamaDecoderLayer
llama_modeling.LlamaModel = MyLlamaModel
#===============================================


class MyLlamaForCausalLM(LlamaForCausalLM):
	def __init__(self, config):
		super().__init__(config)
		self.model.parent_lm_head = self.lm_head #link
		self.num_hidden_layers = config.num_hidden_layers

	def generate(self, **args):
		with torch.no_grad():
			return super().generate(**args)
	
	def clean_layers_weights(self, device="cpu"):
		manifest_map = loader.manifest
		for name, v in manifest_map.items():
			if name.startswith("model.layers."):
				tensor = torch.empty([0], device="cpu")
				#tensor = torch.empty(v["shape"], dtype=torch.float8_e4m3fn, device="cpu") #[0]
				parent, leaf = _walk_to_parent(self, name)
				_assign_tensor_to_module(parent, leaf, tensor)

	def offload_layers_to_cpu(self, layers_num=2):
		for layer_idx in range(min(layers_num, self.num_hidden_layers)):			
			for name, attr in loader.manifest.items():
				if name.startswith(f"model.layers.{layer_idx}."):
					loader.offload_param_to_cpu(name)
		print("./Llama offloading layers to CPU. Done.")



import torch
import math
from typing import Optional

def online_chunked_grouped_attention_rope_no_mask(
    q,                    # (B, Hq, Lq, D)  -- RoPE already applied to q
    k,                    # (B, Hkv, Lk, D) -- RoPE already applied to k
    v,                    # (B, Hkv, Lk, D)
    position_ids: Optional[torch.Tensor] = None,  # (B, L) or (L,) -- not used for masking here
    q_block_size: int = 32768, #64,
    k_block_size: int = 1024, #512,
    eps: float = 1e-12,
):
    """
    Online chunked grouped multi-head attention WITHOUT masking (inference-only).
    - q: (B, Hq, Lq, D) (RoPE already applied)
    - k: (B, Hkv, Lk, D) (RoPE already applied)
    - v: (B, Hkv, Lk, D)
    - position_ids: accepted for API compatibility (came from RoPE) but NOT used here.
    Returns:
      out: (B, Hq, Lq, D)
    Notes:
      * This does streaming softmax (log-sum-exp merge) across k-blocks.
      * Internal accumulators are float32 even if inputs are fp16 for numeric stability.
      * No causal or other masking is applied — it's pure full attention (softmax over all keys).
    """
    assert q.dim() == 4 and k.dim() == 4 and v.dim() == 4
    B, Hq, Lq, D = q.shape
    _, Hkv, Lk, Dk = k.shape
    assert D == Dk and v.shape == (B, Hkv, Lk, D)

    device = q.device
    dtype = q.dtype
    scale = 1.0 / math.sqrt(D)

    # head mapping Hq -> Hkv (bucket each consecutive group to a kv head)
    group_size = (Hq + Hkv - 1) // Hkv 
    head_mapping = torch.arange(Hq, device=device) // group_size  # (Hq,)
    groups = []
    for hkv in range(Hkv):
        q_head_idxs = (head_mapping == hkv).nonzero(as_tuple=False).squeeze(1)
        groups.append(q_head_idxs if q_head_idxs.numel() > 0 else None)

    # position_ids are not used here, but keep signature for compatibility
    # (they should match the RoPE you applied to q/k upstream).
    if position_ids is not None:
        # basic sanity move to device (we don't use it further)
        if position_ids.dim() == 1:
            _ = position_ids.unsqueeze(0).expand(B, -1).to(device)
        else:
            _ = position_ids.to(device)

    out = torch.zeros((B, Hq, Lq, D), device=device, dtype=dtype)

    # iterate over each KV head group (vectorized over batch and group's q-heads)
    for hkv_idx, q_head_idxs in enumerate(groups):
        if q_head_idxs is None:
            continue

        # k_h, v_h: (B, Lk, D) float32
        k_h = k[:, hkv_idx].to(torch.float32)
        v_h = v[:, hkv_idx].to(torch.float32)

        # q_sub: (B, Hq_g, Lq, D) float32
        q_sub = q[:, q_head_idxs].to(torch.float32)
        Hq_g = q_sub.shape[1]

        # process q in q-blocks so we keep accumulators small
        for q_start in range(0, Lq, q_block_size):
            q_end = min(Lq, q_start + q_block_size)
            Bq = q_end - q_start

            q_block = q_sub[:, :, q_start:q_end, :]  # (B, Hq_g, Bq, D)

            # accumulators (float32)
            # m: running max, s: running sum of exp in that frame, wv: running weighted V
            m = torch.full((B, Hq_g, Bq), float("-inf"), device=device, dtype=torch.float32)
            s = torch.zeros((B, Hq_g, Bq), device=device, dtype=torch.float32)
            wv = torch.zeros((B, Hq_g, Bq, D), device=device, dtype=torch.float32)

            # iterate over k-blocks
            for k_start in range(0, Lk, k_block_size):
                k_end = min(Lk, k_start + k_block_size)
                Bk = k_end - k_start

                k_block = k_h[:, k_start:k_end, :]   # (B, Bk, D)
                v_block = v_h[:, k_start:k_end, :]   # (B, Bk, D)

                # scores: (B, Hq_g, Bq, Bk)
                scores = torch.einsum("b h q d, b k d -> b h q k", q_block, k_block) * scale

                # compute local max per row over k
                local_max = torch.amax(scores, dim=-1)  # (B, Hq_g, Bq)
                exp_scores = torch.exp(scores - local_max.unsqueeze(-1))  # (B, Hq_g, Bq, Bk)
                sum_exp = exp_scores.sum(dim=-1)  # (B, Hq_g, Bq)
                weighted_v_chunk = torch.einsum("b h q k, b k d -> b h q d", exp_scores, v_block)  # (B, Hq_g, Bq, D)

                # merge with global accumulators using log-sum-exp merging
                first_chunk_mask = torch.isinf(m)  # True where uninitialized
                # initialize where first_chunk_mask is True
                if first_chunk_mask.any():
                    init_idx = first_chunk_mask
                    m[init_idx] = local_max[init_idx]
                    s[init_idx] = sum_exp[init_idx]
                    wv[init_idx] = weighted_v_chunk[init_idx]

                # merge where already initialized
                merge_idx = ~first_chunk_mask
                if merge_idx.any():
                    m_old = m[merge_idx]
                    s_old = s[merge_idx]
                    wv_old = wv[merge_idx]
                    lm_new = local_max[merge_idx]
                    se_new = sum_exp[merge_idx]
                    wv_new = weighted_v_chunk[merge_idx]

                    new_m = torch.maximum(m_old, lm_new)
                    alpha = torch.exp(m_old - new_m)
                    beta = torch.exp(lm_new - new_m)

                    s[merge_idx] = s_old * alpha + se_new * beta
                    wv[merge_idx] = wv_old * alpha.unsqueeze(-1) + wv_new * beta.unsqueeze(-1)
                    m[merge_idx] = new_m

                # release temporaries
                del scores, local_max, exp_scores, sum_exp, weighted_v_chunk, k_block, v_block

            # finalize output for this q_block
            denom = s.unsqueeze(-1) + eps  # (B, Hq_g, Bq, 1)
            out_block = (wv / denom).to(dtype)  # cast back to original dtype

            # write into output tensor
            out[:, q_head_idxs, q_start:q_end, :] = out_block

            # free accumulators
            del m, s, wv, out_block, denom, q_block

    return out


def online_chunked_grouped_attention_rope(
    q,                    # (B, Hq, Lq, D)
    k,                    # (B, Hkv, Lk, D)
    v,                    # (B, Hkv, Lk, D)
    position_ids: Optional[torch.Tensor] = None,  # (B, L) or (L,)
    causal: bool = True,
    q_block_size: int = 64,
    k_block_size: int = 512,
    eps: float = 1e-12,
):
    """
    Online chunked grouped multi-head attention (RoPE already applied to q,k).
    - q: queries (B, Hq, Lq, D)
    - k,v: keys/values with Hkv heads (B, Hkv, Lk, D)
    - position_ids: used only for causal masking; must match RoPE positions used earlier
    Returns:
      out: (B, Hq, Lq, D)
    """
    assert q.dim() == 4 and k.dim() == 4 and v.dim() == 4
    B, Hq, Lq, D = q.shape
    _, Hkv, Lk, Dk = k.shape
    assert D == Dk and v.shape == (B, Hkv, Lk, D)
    print("attention1.", q.shape, k.shape, v.shape)
    device = q.device
    dtype = q.dtype
    scale = 1.0 / math.sqrt(D)

    # Build head mapping Hq -> Hkv (round-robin / bucket)
    group_size = (Hq + Hkv - 1) // Hkv
    head_mapping = torch.arange(Hq, device=device) // group_size  # (Hq,)
    # For vectorization, collect which q-head indices map to each kv-head
    groups = []
    for hkv in range(Hkv):
        q_head_idxs = (head_mapping == hkv).nonzero(as_tuple=False).squeeze(1)
        if q_head_idxs.numel() == 0:
            groups.append(None)
        else:
            groups.append(q_head_idxs)

    # position ids -> (B, L)
    if position_ids is None:
        pos = torch.arange(Lq, device=device).unsqueeze(0).expand(B, Lq)
    else:
        if position_ids.dim() == 1:
            pos = position_ids.unsqueeze(0).expand(B, -1).to(device)
        else:
            pos = position_ids.to(device)

    out = torch.zeros((B, Hq, Lq, D), device=device, dtype=dtype)

    # We'll iterate over kv-head groups; inside we iterate over q-blocks and k-blocks
    for hkv_idx, q_head_idxs in enumerate(groups):
        if q_head_idxs is None:
            continue

        # slice views for this kv head
        # k_h, v_h: (B, Lk, D)
        k_h = k[:, hkv_idx].to(torch.float32)
        v_h = v[:, hkv_idx].to(torch.float32)

        # q_sub: (B, Hq_g, Lq, D)
        q_sub = q[:, q_head_idxs].to(torch.float32)  # may copy but simpler; small Hq_g usually

        Hq_g = q_sub.shape[1]

        # Process q in blocks (so we don't allocate giant accumulators for full Lq)
        for q_start in range(0, Lq, q_block_size):
            q_end = min(Lq, q_start + q_block_size)
            Bq = q_end - q_start

            q_block = q_sub[:, :, q_start:q_end, :]  # (B, Hq_g, Bq, D)
            q_pos_block = pos[:, q_start:q_end]      # (B, Bq)

            # accumulators per (B, Hq_g, Bq)
            # use float32 accumulators for numeric stability
            m = torch.full((B, Hq_g, Bq), float("-inf"), device=device, dtype=torch.float32)
            s = torch.zeros((B, Hq_g, Bq), device=device, dtype=torch.float32)
            wv = torch.zeros((B, Hq_g, Bq, D), device=device, dtype=torch.float32)

            # iterate over k blocks
            for k_start in range(0, Lk, k_block_size):
                k_end = min(Lk, k_start + k_block_size)
                Bk = k_end - k_start

                k_block = k_h[:, k_start:k_end, :]     # (B, Bk, D) float32
                v_block = v_h[:, k_start:k_end, :]     # (B, Bk, D) float32
                k_pos_block = pos[:, k_start:k_end]    # (B, Bk)

                # scores: (B, Hq_g, Bq, Bk) via einsum
                # q_block: (B, Hq_g, Bq, D); k_block: (B, Bk, D)
                scores = torch.einsum("b h q d, b k d -> b h q k", q_block, k_block) * scale

                # mask for causal: True where masked (k_pos > q_pos)
                if causal:
                    # shape (B, 1, Bq, Bk) broadcastable
                    mask = (k_pos_block.unsqueeze(1).unsqueeze(2) > q_pos_block.unsqueeze(1).unsqueeze(3))
                else:
                    mask = torch.zeros((B, 1, Bq, Bk), dtype=torch.bool, device=device)

                # masked_scores: set masked positions to -inf
                masked_scores = scores.masked_fill(mask, float("-inf"))  # (B, Hq_g, Bq, Bk)

                # local_max per row over k (keep shape (B,Hq_g,Bq))
                local_max = torch.amax(masked_scores, dim=-1)  # (B, Hq_g, Bq)

                # For rows where all positions were masked in this (q_block,k_block) slice,
                # local_max will be -inf; we must handle merging carefully.
                # Compute exp(scores - local_max) but avoid NaN when local_max == -inf by marking invalid rows.
                valid_rows = ~torch.isinf(local_max)  # (B, Hq_g, Bq) boolean

                if valid_rows.any():
                    # For valid rows compute exp and weighted_v
                    # exp_scores shape (B, Hq_g, Bq, Bk)
                    exp_scores = torch.exp(masked_scores - local_max.unsqueeze(-1))
                    sum_exp = exp_scores.sum(dim=-1)  # (B, Hq_g, Bq)
                    # weighted_v_chunk: (B, Hq_g, Bq, D) = einsum over k
                    weighted_v_chunk = torch.einsum("b h q k, b k d -> b h q d", exp_scores, v_block)

                    # Merge streaming softmax results:
                    # If m is -inf (first chunk for that row), initialize. Else merge by log-sum-exp rules.
                    first_chunk_mask = torch.isinf(m)  # (B, Hq_g, Bq) boolean

                    # Where first_chunk_mask is True, initialize m,s,wv
                    init_idx = first_chunk_mask & valid_rows
                    if init_idx.any():
                        # set by indexing
                        m[init_idx] = local_max[init_idx]
                        s[init_idx] = sum_exp[init_idx]
                        wv[init_idx] = weighted_v_chunk[init_idx]

                    # For rows that were already initialized and are valid, merge
                    merge_idx = (~first_chunk_mask) & valid_rows
                    if merge_idx.any():
                        # extract slices
                        m_old = m[merge_idx]                      # (N,)
                        s_old = s[merge_idx]                      # (N,)
                        wv_old = wv[merge_idx]                    # (N, D)
                        lm_new = local_max[merge_idx]             # (N,)
                        se_new = sum_exp[merge_idx]               # (N,)
                        wv_new = weighted_v_chunk[merge_idx]      # (N, D)

                        new_m = torch.maximum(m_old, lm_new)
                        alpha = torch.exp(m_old - new_m)
                        beta = torch.exp(lm_new - new_m)

                        s[merge_idx] = s_old * alpha + se_new * beta
                        # broadcasting alpha,beta to match wv shapes
                        wv[merge_idx] = wv_old * alpha.unsqueeze(-1) + wv_new * beta.unsqueeze(-1)
                        m[merge_idx] = new_m

                    # free temporaries (helpful when memory is tight)
                    del exp_scores, sum_exp, weighted_v_chunk

                # else: if no valid rows in this slice, nothing to merge

                # free per-block tensors we no longer need
                del scores, masked_scores, local_max, mask, k_block, v_block, k_pos_block

            # after all k-blocks: finalize output for this q_block
            # denom (B, Hq_g, Bq, 1)
            denom = s.unsqueeze(-1) + eps
            out_block = (wv / denom).to(dtype)  # cast back to original dtype

            # write into out: out[:, q_head_idxs, q_start:q_end, :]
            # out_block shape (B, Hq_g, Bq, D)
            out[:, q_head_idxs, q_start:q_end, :] = out_block

            # free accumulators for this q_block
            del m, s, wv, out_block, denom, q_block

    return out



#=== Try torch.compile later reduce overhead ===
"""
# Option A: balanced / safe
compiled_attn = torch.compile(
    online_chunked_grouped_attention_rope,
    backend="inductor",      # default; uses triton + generated kernels on GPU
    mode="reduce-overhead"   # good for inference: lower overhead, uses CUDA graphs
)

# Option B: maximum steady-state speed (long compile time; autotunes GEMMs)
compiled_attn_max = torch.compile(
    online_chunked_grouped_attention_rope,
    backend="inductor",
    mode="max-autotune"      # autotunes GEMM backends; long first-run compile
)

# Call it exactly like the original call
out = compiled_attn(q, k, v, position_ids=pos, causal=True, q_block_size=64, k_block_size=512)
"""

if __name__=="__main__":
    # -------------------------
    # Test with your sample
    # -------------------------
    B, Hq, Hkv, L, D = 2, 16, 4, 256, 64
    q = torch.randn(B, Hq, L, D, device="cuda")
    k = torch.randn(B, Hkv, L, D, device="cuda")
    v = torch.randn(B, Hkv, L, D, device="cuda")

    out = online_chunked_grouped_attention_rope(q, k, v, q_block_size=64, k_block_size=128)
    print(out.shape)  # should be (2,16,256,64)



"""FlashAttention w/support for learned sinks and banded attention.

This is an expanded version of the Flash Attention v2 implementation (see https://tridao.me/publications/flash2/flash2.pdf)
which can be found at https://triton-lang.org/main/getting-started/tutorials/06-fused-attention.html.

This version has been extended to support banded attention and learned attention sinks.
"""

#import pytest
import torch

import triton
import triton.language as tl
from triton.tools.tensor_descriptor import TensorDescriptor


@triton.jit
def _attn_fwd(
    Q,
    K,
    V,
    Sinks,
    sm_scale,
    M,
    Out,  #
    Start_q,
    Z,
    H,
    N_Q_CTX,
    N_KV_CTX,
    HEAD_DIM: tl.constexpr,  #
    BLOCK_M: tl.constexpr,  #
    BLOCK_N: tl.constexpr,  #
    BANDWIDTH: tl.constexpr,
):
    tl.static_assert(BLOCK_N <= HEAD_DIM)
    start_q = tl.load(Start_q).to(tl.int32)
    start_m = tl.program_id(0)
    off_hz = tl.program_id(1)
    off_z = off_hz // H
    off_h = off_hz % H

    # load attention sinks
    if Sinks is not None:
        sink = tl.load(Sinks + off_h).to(tl.float32)
    else:
        sink = 0

    # initialize offsets
    offs_m = start_m * BLOCK_M + tl.arange(0, BLOCK_M)
    offs_n = tl.arange(0, BLOCK_N)
    # initialize pointer to m and l
    m_i = tl.zeros([BLOCK_M], dtype=tl.float32) + sink
    l_i = tl.zeros([BLOCK_M], dtype=tl.float32)
    acc = tl.zeros([BLOCK_M, HEAD_DIM], dtype=tl.float32)
    # load scales
    qk_scale = sm_scale
    q = Q.load([off_z, off_h, start_m * BLOCK_M, 0]).reshape([BLOCK_M, HEAD_DIM])

    if BANDWIDTH:
        lo, hi = tl.maximum(start_q, start_q + start_m * BLOCK_M - BANDWIDTH), start_q + (start_m + 1) * BLOCK_M
    else:
        lo, hi = start_q, start_q + (start_m + 1) * BLOCK_M

    for start_n in range(lo, hi, BLOCK_N):
        start_n = tl.multiple_of(start_n, BLOCK_N)

        mask = (start_n + offs_n)[None, :] > (start_q + offs_m)[:, None]

        if BANDWIDTH:
            too_old = (start_n + offs_n[None, :]) < (start_q + offs_m[:, None] - BANDWIDTH + 1)
            mask = mask | too_old

        k = K.load([off_z, off_h, start_n, 0]).reshape([BLOCK_N, HEAD_DIM]).T
        qk = tl.dot(q, k, allow_tf32=False)

        qk = qk * qk_scale + tl.where(mask, -1.0e6, 0.0)
        m_ij = tl.maximum(m_i, tl.max(qk, 1))
        qk -= m_ij[:, None]

        p = tl.math.exp(qk)
        alpha = tl.math.exp(m_i - m_ij)
        l_ij = tl.sum(p, 1)
        acc = acc * alpha[:, None]

        v = V.load([off_z, off_h, start_n, 0]).reshape([BLOCK_N, HEAD_DIM])
        v = v.to(tl.float32)
        acc = tl.dot(p, v, acc, allow_tf32=False)

        l_i = l_i * alpha + l_ij
        m_i = m_ij

    sink = tl.math.exp(sink - m_i)
    z = l_i + sink
    acc = acc / z[:, None]
    m_i += tl.math.log(l_i)
    m_ptrs = M + off_hz * N_Q_CTX + offs_m
    tl.store(m_ptrs, m_i)
    acc = acc.to(Out.dtype)[None, None, :, :]
    Out.store([off_z, off_h, start_m * BLOCK_M, 0], acc)


class _attention(torch.autograd.Function):
    @staticmethod
    def forward(ctx, q, k, v, sinks, sm_scale, bandwidth, start_q):
        assert len(start_q) == 1
        """
        bs, n_ctx, n_kv_heads, repeat_kv, HEAD_DIM_Q = q.shape
        bs, n_kv_ctx, n_kv_heads, HEAD_DIM_K = k.shape
        bs, n_kv_ctx, n_kv_heads, HEAD_DIM_V = v.shape
        n_heads = n_kv_heads * repeat_kv
        q = q.view(bs, n_ctx, n_heads, HEAD_DIM_Q)
        k = k.view(bs, n_kv_ctx, n_kv_heads, HEAD_DIM_K)
        assert HEAD_DIM_Q == HEAD_DIM_K and HEAD_DIM_K == HEAD_DIM_V
        assert HEAD_DIM_K in {16, 32, 64, 128, 256}

        q = q.transpose(1, 2).contiguous()
        k = k.repeat_interleave(repeat_kv, dim=2).transpose(1, 2).contiguous()
        v = v.repeat_interleave(repeat_kv, dim=2).transpose(1, 2).contiguous()
        """
        #meine
        bs, n_ctx, n_kv_ctx, n_heads, HEAD_DIM_K, HEAD_DIM_V = q.shape[0], q.shape[2], k.shape[2], 64, 64, 64
        #print("n_ctx:", n_ctx, "n_kv_ctx:", n_kv_ctx)
        #=====

        BLOCK_M = 64
        BLOCK_N = 64
        m_pad_size = BLOCK_M - n_ctx % BLOCK_M if n_ctx % BLOCK_M != 0 else 0
        # pad q to multiple of its block size in the n_ctx dimension (-2)
        q = torch.nn.functional.pad(q, (0, 0, 0, m_pad_size))
        n_pad_size = BLOCK_N - n_kv_ctx % BLOCK_N if n_kv_ctx % BLOCK_N != 0 else 0
        # pad k and v to multiple of their block size in the n_kv_ctx dimension
        k = torch.nn.functional.pad(k, (0, 0, 0, n_pad_size))
        v = torch.nn.functional.pad(v, (0, 0, 0, n_pad_size))

        o = torch.empty_like(q)
        M = torch.empty((bs, n_heads, n_ctx + m_pad_size), device=q.device, dtype=torch.float32)
        grid = (triton.cdiv(n_ctx, BLOCK_M), bs * n_heads, 1)
        _attn_fwd[grid](
            TensorDescriptor.from_tensor(q, [1, 1, BLOCK_M, HEAD_DIM_K]),
            TensorDescriptor.from_tensor(k, [1, 1, BLOCK_N, HEAD_DIM_K]),
            TensorDescriptor.from_tensor(v, [1, 1, BLOCK_N, HEAD_DIM_K]),
            sinks,
            sm_scale,
            M,
            TensorDescriptor.from_tensor(o, [1, 1, BLOCK_M, HEAD_DIM_K]),
            start_q,
            q.shape[0],
            q.shape[1],
            N_Q_CTX=n_ctx + m_pad_size,
            N_KV_CTX=n_kv_ctx,
            HEAD_DIM=HEAD_DIM_K,
            BANDWIDTH=bandwidth,
            BLOCK_M=BLOCK_M,
            BLOCK_N=BLOCK_N,
        )

        ctx.save_for_backward(q, k, v, sinks, o, M, start_q)
        ctx.sm_scale = sm_scale
        ctx.bandwidth = bandwidth

        o = o[:, :, :n_ctx, :].transpose(1, 2).contiguous()
        #o = o.view(bs, n_ctx, n_heads * HEAD_DIM_V) #will make it in AttentionBlock
        return o


attention = _attention.apply


def attention_ref(
    query: torch.Tensor,
    key: torch.Tensor,
    value: torch.Tensor,
    sinks: torch.Tensor,
    sm_scale: float = 0.125,
    sliding_window: int | None = None,
    start_q: torch.LongTensor = 0,
):
    batch_size, num_queries, num_key_value_heads, num_key_value_groups, head_dim = query.shape
    batch_size, num_keys, num_key_value_heads, head_dim = key.shape

    sinks = sinks.view(1, num_key_value_heads, num_key_value_groups, 1, 1).float()
    key = key.unsqueeze(3)
    value = value.unsqueeze(3)

    pos_keys = torch.arange(num_keys, device=query.device)
    pos_queries = torch.arange(num_queries, device=query.device) + start_q
    mask = pos_keys[None, :] > pos_queries[:, None]
    mask = mask.float().masked_fill(mask, float("-inf"))

    if sliding_window:
        too_old = pos_keys[None, :] < (pos_queries[:, None] - sliding_window + 1)
        mask.masked_fill_(too_old, float("-inf"))

    logits = torch.einsum("bqhmd,bkhmd->bhmqk", query.float(), key.float()) * sm_scale
    logits = logits + mask[None, None, None, :, :]

    logits_max = torch.max(logits, dim=-1, keepdim=True).values
    logits_or_sinks_max = torch.maximum(sinks, logits_max)
    sinks = torch.exp(sinks - logits_or_sinks_max)
    unnormalized_scores = torch.exp(logits - logits_or_sinks_max)
    normalizer = unnormalized_scores.sum(dim=-1, keepdim=True) + sinks
    scores = unnormalized_scores / normalizer

    output = torch.einsum("bhmqk,bkhmd->bqhmd", scores, value.float())

    output = output.reshape(batch_size, num_queries, num_key_value_heads * num_key_value_groups * head_dim).bfloat16()
    return output

"""
@pytest.mark.parametrize("batch_size", [1, 2])
@pytest.mark.parametrize("num_queries", [1, 128])
@pytest.mark.parametrize("num_keys", [128, 32])
@pytest.mark.parametrize("num_key_value_heads", [8])
@pytest.mark.parametrize("num_key_value_groups", [8])
@pytest.mark.parametrize("head_dim", [64])
@pytest.mark.parametrize("sm_scale", [0.125])
@pytest.mark.parametrize("sliding_window", [None, 128])
@pytest.mark.parametrize("start_q", [0, 5])
"""
def test_eq(batch_size, num_queries, num_keys, num_key_value_heads, num_key_value_groups, head_dim, sm_scale, sliding_window, start_q):
    if num_queries > num_keys:
        pytest.skip("too many queries")

    q = torch.randn(batch_size, num_queries, num_key_value_heads, num_key_value_groups, head_dim).bfloat16().cuda()
    k = torch.randn(batch_size, num_keys, num_key_value_heads, head_dim).bfloat16().cuda()
    v = torch.randn(batch_size, num_keys, num_key_value_heads, head_dim).bfloat16().cuda()
    sinks = torch.randn(num_key_value_heads * num_key_value_groups).bfloat16().cuda()

    start_q = torch.tensor([start_q], dtype=torch.int32).cuda()

    o1 = attention(q, k, v, sinks, sm_scale, sliding_window, start_q)
    o2 = attention_ref(q, k, v, sinks, sm_scale, sliding_window, start_q)

    torch.testing.assert_close(o1, o2)


import codecs, io, time
import torch 

def file_put_contents(filename, st):
    file = codecs.open(filename, "w", "utf-8")
    file.write(st)
    file.close()

def file_get_contents(name):
    f = io.open(name, mode="r", encoding="utf-8") #utf-8 | Windows-1252
    return f.read()
    
def tensor_size_gb(t: torch.Tensor) -> float:
    return t.numel() * t.element_size() / 1024**3

class Stats:
    def __init__(self):
        self.d = {}

    def set(self, name, t1):
        if name not in self.d: self.d[name] = []
        self.d[name].append( round(time.perf_counter() - t1, 3) ) 

    def print_and_clean(self):
        st = "Stats:"
        for name, a in self.d.items():
            st+=f" {name}: {a[:5]} t:{round(sum(a), 3)},"
        self.d = {}
        return st


# === Helper utilities ===
def _walk_to_parent(obj, attr_path):
    """Return (parent_obj, leaf_name) for attr_path like 'self_attn.q_proj.weight'"""
    parts = attr_path.split('.')
    parent = obj
    for p in parts[:-1]:
        parent = getattr(parent, p)
    return parent, parts[-1]

def _assign_tensor_to_module(target_parent, leaf, tensor):
    """
    Assign a tensor into target_parent.<leaf>.
    - If target_parent.<leaf> has a .load call, call it with tensor.
    - Else, if attribute endswith 'weight' or 'bias' and current attr is nn.Parameter, replace it.
    - Else, set attribute to nn.Parameter(tensor) (read-only).
    """
    existing = getattr(target_parent, leaf, None)

    # If target object has a load(tensor) method (user's custom modules), call it.
    if hasattr(existing, "load") and callable(getattr(existing, "load")):
        existing.load(tensor)   # user-supplied API
        return

    # If existing is a Parameter (typical), replace with new Parameter on CUDA
    if isinstance(existing, torch.nn.Parameter) or getattr(existing, "__class__", None) is torch.nn.Parameter:
        param = torch.nn.Parameter(tensor.detach(), requires_grad=False)
        setattr(target_parent, leaf, param)
        return

    # If attribute is a module (like a Linear) we attempt to set its weight/bias
    if isinstance(existing, torch.nn.Linear) or hasattr(existing, "weight"):
        # try to set weight and bias if given tensor is 2D weight
        if tensor.ndim == 2 and hasattr(existing, "weight"):
            existing.weight = torch.nn.Parameter(tensor.detach(), requires_grad=False)
            return
        # fallback: set attribute to Parameter
    # Default fallback: replace attribute with a Parameter
    setattr(target_parent, leaf, torch.nn.Parameter(tensor.detach(), requires_grad=False))


def _set_meta_placeholder(target_parent, leaf):
    """Replace parameter/module attribute with a tiny meta-device Parameter to free VRAM."""
    placeholder = torch.nn.Parameter(torch.empty(0, device="meta"), requires_grad=False)
    setattr(target_parent, leaf, placeholder)


def remove_layers_weights(model):
    # 2. Remove heavy decoder block weights (keep skeleton)
    for layer in model.model.layers:
        for name, module in layer.named_children():
            if hasattr(module, "weight"):
                module.weight = torch.nn.Parameter(
                    torch.empty(0), requires_grad=False
                )                
            if hasattr(module, "bias") and module.bias is not None:
                module.bias = torch.nn.Parameter(
                    torch.empty(0), requires_grad=False
                )





# efficiant gpt-oss-20B that runs on consumer PC with 8GB VRAM

import time, os, math
from datetime import datetime
import numpy as np
import torch
from torch import nn
import torch.nn.functional as F
from typing import Callable, Optional, Tuple, Union, Dict, Any, Iterable, List, Unpack
from transformers import GptOssForCausalLM, AutoTokenizer, AutoModelForCausalLM
from .utils import _walk_to_parent, _assign_tensor_to_module, _set_meta_placeholder
from .gds_loader import GDSWeights
from .gpt_oss_attention import attention as flash_attention

#global vars
loader, stats = None, None

#======== rewriting core classess ==============
from transformers.models.gpt_oss.modeling_gpt_oss import GptOssAttention, GptOssExperts,  GptOssModel, GptOssConfig, GptOssDecoderLayer, create_causal_mask, create_sliding_window_causal_mask, repeat_kv, MoeModelOutputWithPast

class MyGptOssAttention(GptOssAttention):
	def forward(self, *args, **kwargs):
		out = super().forward(*args, **kwargs)
		#print(self.layer_idx, "attention:", out[0].shape)
		return out


class MyGptOssExperts(GptOssExperts):
	def forward(self, hidden_states: torch.Tensor, router_indices=None, routing_weights=None) -> torch.Tensor: #chunked/looped
		num_experts = routing_weights.shape[1]
		batch_size, T, H = hidden_states.shape
		acc = torch.zeros((batch_size, T, H), device=hidden_states.device, dtype=hidden_states.dtype)
		for e in range(num_experts):
			s_gate_up_proj = self.gate_up_proj[e].unsqueeze(0)
			s_gate_up_proj_bias = self.gate_up_proj_bias[e].unsqueeze(0)
			s_down_proj = self.down_proj[e].unsqueeze(0)
			s_down_proj_bias = self.down_proj_bias[e].unsqueeze(0)
			routing_weights_e = routing_weights[:, e].unsqueeze(-1)

			gate_up = torch.bmm(hidden_states, s_gate_up_proj) + s_gate_up_proj_bias[..., None, :]
			gate, up = gate_up[..., ::2], gate_up[..., 1::2]
			gate = gate.clamp(min=None, max=self.limit)
			up = up.clamp(min=-self.limit, max=self.limit)
			glu = gate * torch.sigmoid(gate * self.alpha)
			next_states = torch.bmm(((up + 1) * glu), s_down_proj)
			next_states = next_states + s_down_proj_bias[..., None, :]
			next_states = next_states.view(1, batch_size, -1, self.hidden_size)
			next_states = next_states * routing_weights_e.transpose(0, 1).view(1, batch_size, -1)[..., None]
			next_states = next_states.squeeze(0) #.sum(dim=0)
			acc += next_states
			#print(x.shape, s_gate_up_proj.shape, s_gate_up_proj_bias.shape, s_down_proj.shape, "routing_weights:", routing_weights_e.shape, "next_states:", next_states.shape)			
		return acc


	def forward_default(self, hidden_states: torch.Tensor, router_indices=None, routing_weights=None) -> torch.Tensor:
		#t1 = self.forward_chunked(hidden_states, routing_weights=routing_weights) #chunked for comparison
		
		batch_size = hidden_states.shape[0]
		hidden_states = hidden_states.reshape(-1, self.hidden_size)  # (num_tokens, hidden_size)
		num_experts = routing_weights.shape[1]
					
		hidden_states = hidden_states.repeat(num_experts, 1) #T,C->num_experts,T,C
		hidden_states = hidden_states.view(num_experts, -1, self.hidden_size)

		gate_up = torch.bmm(hidden_states, self.gate_up_proj) + self.gate_up_proj_bias[..., None, :]
		gate, up = gate_up[..., ::2], gate_up[..., 1::2]
		gate = gate.clamp(min=None, max=self.limit)
		up = up.clamp(min=-self.limit, max=self.limit)
		glu = gate * torch.sigmoid(gate * self.alpha)
		next_states = torch.bmm(((up + 1) * glu), self.down_proj)
		next_states = next_states + self.down_proj_bias[..., None, :]
		next_states = next_states.view(num_experts, batch_size, -1, self.hidden_size)
		next_states = next_states * routing_weights.transpose(0, 1).view(num_experts, batch_size, -1)[..., None]
		next_states = next_states.sum(dim=0)
		#print(t1.flatten()[-5:], next_states.flatten()[-5:], "maxErr:", (t1-next_states).abs().max().item(), "\n")
		return next_states



class oDecoderLayer:
	def _get_my_manifests(self):
		a = []
		for manifest_name in loader.manifest.keys():
			base = f"model.layers.{self.layer_idx}."
			if not manifest_name.startswith(base): continue
			attr_path = manifest_name.replace(base, "")
			a.append((manifest_name, attr_path))
		return a

	def _load_layer_weights(self):
		for manifest_name, attr_path in self._get_my_manifests():
			try:
				t1 = time.perf_counter()
				tensor = loader.load_param_to_cuda(manifest_name)
				parent, leaf = _walk_to_parent(self, attr_path)
				_assign_tensor_to_module(parent, leaf, tensor)
				if stats: stats.set("layer_load", t1)
			except Exception as e:
				raise RuntimeError(f"failed to load {manifest_name} into {attr_path}: {e}")

	def _unload_layer_weights(self):
		for manifest_name, attr_path in self._get_my_manifests():
			parent, leaf = _walk_to_parent(self, attr_path)
			_set_meta_placeholder(parent, leaf)


class MyGptOssDecoderLayer(GptOssDecoderLayer, oDecoderLayer):
	def __init__(self, config: GptOssConfig, layer_idx: int):
		super().__init__(config, layer_idx)	
		self.layer_idx = layer_idx

	def forward(self, *args, **kwargs):
		self._load_layer_weights()
		out = super().forward(*args, **kwargs)
		self._unload_layer_weights()
		return out


class MyGptOssModel(GptOssModel):
	def __init__(self, config: GptOssConfig):		
		super().__init__(config)
		self.config = config
		self.layers = nn.ModuleList([MyGptOssDecoderLayer(config, layer_idx) for layer_idx in range(2)])
		for decoder_layer in self.layers: decoder_layer._unload_layer_weights()		

	def forward(
		self,
		input_ids: Optional[torch.LongTensor] = None,
		attention_mask: Optional[torch.Tensor] = None,
		position_ids: Optional[torch.LongTensor] = None,
		past_key_values: Optional[list[torch.FloatTensor]] = None,
		inputs_embeds: Optional[torch.FloatTensor] = None,
		use_cache: Optional[bool] = None,
		cache_position: Optional[torch.LongTensor] = None,
		**kwargs: Unpack,
	) -> MoeModelOutputWithPast:
		if (input_ids is None) ^ (inputs_embeds is not None):
			raise ValueError("You must specify exactly one of input_ids or inputs_embeds")

		if use_cache and past_key_values is None:
			past_key_values = DynamicCache()

		if inputs_embeds is None:
			inputs_embeds = self.embed_tokens(input_ids)

		if cache_position is None:
			past_seen_tokens = past_key_values.get_seq_length() if past_key_values is not None else 0
			cache_position = torch.arange(
				past_seen_tokens, past_seen_tokens + inputs_embeds.shape[1], device=inputs_embeds.device
			)
		if position_ids is None:
			position_ids = cache_position.unsqueeze(0)

		# It may already have been prepared by e.g. `generate`
		if not isinstance(causal_mask_mapping := attention_mask, dict):
			mask_kwargs = {
				"config": self.config,
				"input_embeds": inputs_embeds,
				"attention_mask": attention_mask,
				"cache_position": cache_position,
				"past_key_values": past_key_values,
			}
			causal_mask_mapping = {
				"full_attention": create_causal_mask(**mask_kwargs),
				"sliding_attention": create_sliding_window_causal_mask(**mask_kwargs),
			}

		hidden_states = inputs_embeds
		position_embeddings = self.rotary_emb(hidden_states, position_ids)

		# ===== meine ========================
		self.embed_tokens.cpu(); self.parent_lm_head.cpu()

		for layer_idx in range(self.config.num_hidden_layers):
			decoder_layer = self.layers[layer_idx % 2]
			decoder_layer.layer_idx = layer_idx
			decoder_layer.self_attn.layer_idx = layer_idx
			hidden_states = decoder_layer(
				hidden_states,
				attention_mask=causal_mask_mapping[decoder_layer.attention_type],
				position_ids=position_ids,
				past_key_value=past_key_values,
				use_cache=use_cache,
				cache_position=cache_position,
				position_embeddings=position_embeddings,
				**kwargs,
			)

		if stats: print("./gpt_oss.forward.", datetime.now().strftime("%H:%M:%S"), stats.print_and_clean() if stats else "")
		hidden_states = self.norm(hidden_states)
		self.embed_tokens.to(hidden_states.device); self.parent_lm_head.to(hidden_states.device)
		#./===================================
		
		return MoeModelOutputWithPast(
			last_hidden_state=hidden_states,
			past_key_values=past_key_values,
		)
		

#===============================================
def my_eager_attention_forward(
	module: nn.Module,
	query: torch.Tensor,
	key: torch.Tensor,
	value: torch.Tensor,
	attention_mask: Optional[torch.Tensor],
	scaling: float,
	dropout: float = 0.0,
	sliding_window=None,
	s_aux = None,
	**kwargs,
):
	key_states = repeat_kv(key, module.num_key_value_groups)
	value_states = repeat_kv(value, module.num_key_value_groups)	
	
	# Flash-attention
	offset, n_ctx = min(key.shape[2] - query.shape[2], sliding_window if sliding_window else 999), query.shape[2]
	if offset==0: #use FA only for first generation
		#print("offset", query.shape, key.shape, offset, "n_ctx:", n_ctx, "sliding_window:", sliding_window, "scaling:", scaling, kwargs)
		start_q = torch.LongTensor([offset]).to(query.device)
		t = flash_attention(
			query,
			key_states,
			value_states,
			s_aux, #sinks,
			scaling,
			sliding_window,
			start_q
		)
		attn_output, attn_weights = t, None
	
	else: #standard attention (default)
		attn_weights = torch.matmul(query, key_states.transpose(2, 3)) * scaling
		if attention_mask is not None:
			causal_mask = attention_mask[:, :, :, : key_states.shape[-2]]
			attn_weights = attn_weights + causal_mask

		sinks = module.sinks.reshape(1, -1, 1, 1).expand(query.shape[0], -1, query.shape[-2], -1)
		combined_logits = torch.cat([attn_weights, sinks], dim=-1)

		# This was not in the original implementation and slightly affect results; it prevents overflow in BF16/FP16
		# when training with bsz>1 we clamp max values.
		combined_logits = combined_logits - combined_logits.max(dim=-1, keepdim=True).values
		probs = F.softmax(combined_logits, dim=-1, dtype=combined_logits.dtype)
		scores = probs[..., :-1]  # we drop the sink here
		attn_weights = nn.functional.dropout(scores, p=dropout, training=module.training)
		attn_output = torch.matmul(attn_weights, value_states)	
		attn_output = attn_output.transpose(1, 2).contiguous()
	
	#print("my_eager_attention_forward:", attn_output.shape, t.shape, attn_output.flatten()[-5:], t.flatten()[-5:], "Error:", (attn_output - t).abs().max().item(), "\n")
	return attn_output, attn_weights


import transformers.models.gpt_oss.modeling_gpt_oss as modeling
modeling.GptOssExperts = MyGptOssExperts
modeling.GptOssAttention = MyGptOssAttention
modeling.GptOssModel = MyGptOssModel
modeling.eager_attention_forward = my_eager_attention_forward
#===============================================


class MyGptOssForCausalLM(GptOssForCausalLM):
	def __init__(self, config):
		super().__init__(config)
		self.model.parent_lm_head = self.lm_head #link
		self.num_hidden_layers = config.num_hidden_layers

	def generate(self, **args):
		with torch.no_grad():
			return super().generate(**args)

	def offload_layers_to_cpu(self, layers_num=2):
		layer = oDecoderLayer()
		for layer_idx in range(min(layers_num, self.num_hidden_layers)):
			layer.layer_idx = layer_idx
			for manifest_name, attr_path in layer._get_my_manifests():
				loader.offload_param_to_cpu(manifest_name)
		print("./gpt_oss offloading layers to CPU. Done.")



# 4.57.0.dev qwen3_next

import time, os, math, json
from datetime import datetime
import numpy as np
import torch
from torch import nn
import torch.nn.functional as F
from typing import Callable, Optional, Tuple, Union, Dict, Any, Iterable, List, Unpack
from .utils import _walk_to_parent, _assign_tensor_to_module, _set_meta_placeholder, file_get_contents
from .kvcache import oCache

#global vars
loader, stats = None, None

#======== rewriting core classes ==============
from transformers.models.qwen3_next.modeling_qwen3_next import Qwen3NextMLP, Qwen3NextSparseMoeBlock, Qwen3NextDecoderLayer, Qwen3NextConfig, Qwen3NextModel, Qwen3NextForCausalLM, Qwen3NextDynamicCache, Qwen3NextRMSNorm, create_causal_mask, repeat_kv, MoeModelOutputWithPast, MoeCausalLMOutputWithPast, TransformersKwargs, Cache

class Qwen3NextDiskCache(Qwen3NextDynamicCache, oCache):
	def __init__(self, config, cache_dir="./kv_cache", stats=None):
		super().__init__(config)
		self.ini_ocache(cache_dir, stats)
		self.seq_lengths = [0 for _ in range(len(self.key_cache))]

	def get_seq_length(self, layer_idx: Optional[int] = 0) -> int:
		return self.seq_lengths[layer_idx]

	def __getitem__(self, layer_idx: int) -> tuple[torch.Tensor, torch.Tensor]:
		raise NotImplementedError("KVCache __getitem__ called. Beam search is not supported")

	def reorder_cache(self, beam_idx: torch.LongTensor):
		raise NotImplementedError("KVCache reorder_cache called. Beam search is not supported")

	def update(
		self,
		key_states: torch.Tensor,
		value_states: torch.Tensor,
		layer_idx: int,
		cache_kwargs: Optional[Dict[str, Any]] = None,
	) -> Tuple[torch.Tensor, torch.Tensor]:
		tensors = self.load_from_disk(layer_idx)
		if tensors is not None:
			self.key_cache[layer_idx], self.value_cache[layer_idx] = tensors
			if layer_idx < len(self.key_cache2):
				self.key_cache[layer_idx] = torch.cat([self.key_cache[layer_idx], self.key_cache2[layer_idx]], dim=-2)
				self.value_cache[layer_idx] = torch.cat([self.value_cache[layer_idx], self.value_cache2[layer_idx]], dim=-2)
				self.key_cache2[layer_idx] = torch.cat([self.key_cache2[layer_idx], key_states], dim=-2)
				self.value_cache2[layer_idx] = torch.cat([self.value_cache2[layer_idx], value_states], dim=-2)
			else:
				self.key_cache2.append(key_states)
				self.value_cache2.append(value_states)
		
		out = super().update(key_states, value_states, layer_idx, cache_kwargs) #tuple of (self.key_cache[layer_idx], self.value_cache[layer_idx])
		self.seq_lengths[layer_idx] = out[0].shape[-2]
		#print(len(out), out[0].shape, "-- k shape" )
		if tensors is None: self.save_to_disk(out, layer_idx) #save only first time cause it's slow to save
		self.key_cache[layer_idx], self.value_cache[layer_idx] = torch.empty(0), torch.empty(0)
		return out


class loaderLayer:
	def _load_layer_weights(self):
		t1 = time.perf_counter()
		base = f"model.layers.{self.layer_idx}."
		loader.preload_layer_safetensors(base)
		d = loader.load_dict_to_cuda(base)
		for attr_path, tensor in d.items():
			parent, leaf = _walk_to_parent(self, attr_path)
			_assign_tensor_to_module(parent, leaf, tensor)
		if stats: stats.set("layer_load", t1)
			
	def _unload_layer_weights(self):
		base = f"model.layers.{self.layer_idx}."
		for attr_path in loader.manifest[base]:
			parent, leaf = _walk_to_parent(self, attr_path)
			_set_meta_placeholder(parent, leaf)

	def _load_expert_weights(self):
		t1 = time.perf_counter()
		base = f"model.layers.{self.layer_idx}.mlp.experts.{self.expert_idx}."
		d = loader.load_dict_to_cuda(base)
		for attr_path, tensor in d.items():
			parent, leaf = _walk_to_parent(self, attr_path)
			_assign_tensor_to_module(parent, leaf, tensor)
		if stats: stats.set("experts_load", t1)

	def _unload_expert_weights(self):
		base = f"model.layers.{self.layer_idx}.mlp.experts.{self.expert_idx}."
		for attr_path in loader.manifest[base]:
			parent, leaf = _walk_to_parent(self, attr_path)
			_set_meta_placeholder(parent, leaf)

	def _load_experts_weights(self, experts_idx):
		for expert_idx in experts_idx:
			self.experts[expert_idx]._load_expert_weights()

	def _unload_experts_weights(self, experts_idx):
		for expert_idx in experts_idx:
			self.experts[expert_idx]._unload_expert_weights()


class MyQwen3NextMLP(Qwen3NextMLP, loaderLayer):
	def forward(self, x):
		if hasattr(self, "expert_idx"): self._load_expert_weights()
		out = super().forward(x)
		if hasattr(self, "expert_idx"): self._unload_expert_weights()
		return out
		

class MyQwen3NextSparseMoeBlock(Qwen3NextSparseMoeBlock, loaderLayer):
	def forward_stacked(self, hidden_states: torch.Tensor) -> torch.Tensor: #stacked+chunked experimental
		batch_size, sequence_length, hidden_dim = hidden_states.shape
		hidden_states = hidden_states.view(-1, hidden_dim)
		# router_logits: (batch * sequence_length, n_experts)
		router_logits = self.gate(hidden_states)

		routing_weights = F.softmax(router_logits, dim=1, dtype=torch.float)
		routing_weights, selected_experts = torch.topk(routing_weights, self.top_k, dim=-1)
		if self.norm_topk_prob:
			routing_weights /= routing_weights.sum(dim=-1, keepdim=True)
		# we cast back to the input dtype
		routing_weights = routing_weights.to(hidden_states.dtype)

		final_hidden_states = torch.zeros(
			(batch_size * sequence_length, hidden_dim), dtype=hidden_states.dtype, device=hidden_states.device
		)

		# One hot encode the selected experts to create an expert mask
		# this will be used to easily index which expert is going to be sollicitated
		expert_mask = torch.nn.functional.one_hot(selected_experts, num_classes=self.num_experts).permute(2, 1, 0)

		# Loop over all available experts in the model and perform the computation on each expert
		expert_hit = torch.greater(expert_mask.sum(dim=(-1, -2)), 0).nonzero()
		"""
		for expert_idx in expert_hit:
			expert_layer = self.experts[expert_idx]
			idx, top_x = torch.where(expert_mask[expert_idx].squeeze(0))
			current_state = hidden_states[None, top_x].reshape(-1, hidden_dim)
			current_hidden_states = expert_layer(current_state) * routing_weights[top_x, idx, None]
			final_hidden_states.index_add_(0, top_x, current_hidden_states.to(hidden_states.dtype))	
		"""		
		#===== Chunked Stacked part =====================================
		#for start_idx in range(0, experts_hit.shape[0], 5):
		#expert_hit = experts_hit[start_idx : start_idx+5]

		# assume experts all share dimensions:
		# in_dim -> hidden_dim -> out_dim
		token_pos_list, expert_id_list, local_rank_list = [], [], []

		for expert_idx in expert_hit:
			local_rank, tok_pos = torch.where(expert_mask[expert_idx].squeeze(0))
			if tok_pos.numel() == 0: continue #?
			token_pos_list.append(tok_pos)
			expert_id_list.append(torch.full_like(tok_pos, int(expert_idx)))
			local_rank_list.append(local_rank)
		
		token_pos = torch.cat(token_pos_list)      # (M,)
		expert_ids = torch.cat(expert_id_list)     # (M,)
		local_rank = torch.cat(local_rank_list)    # (M,)
		print(expert_hit.shape, expert_ids.shape)

		x = hidden_states[token_pos]               # (M, in_dim)
		M, in_dim = x.shape
		device, dtype = x.device, x.dtype

		# ---- Gather expert params as stacked tensors ----
		def stack_params(attr):
			return torch.stack([getattr(self.experts[eid], attr).weight for eid in expert_ids], dim=0)

		#expert_indexes = expert_hit.cpu().squeeze().tolist()
		self._load_experts_weights(expert_hit)
		W_gate = stack_params("gate_proj")     # (E, hidden, in), (E, hidden)
		W_up   = stack_params("up_proj")       # (E, hidden, in), (E, hidden)
		W_down = stack_params("down_proj")     # (E, out, hidden), (E, out)
		self._unload_experts_weights(expert_hit)

		# slice only the experts actually used by these tokens
		#W_gate = W_gate[expert_ids].to(device=device, dtype=dtype)  # (M, hidden, in)
		#W_up   = W_up[expert_ids].to(device=device, dtype=dtype)		
		#W_down = W_down[expert_ids].to(device=device, dtype=dtype)				

		# ---- Layer by layer ----
		x_un = x.unsqueeze(-1)                       # (M, in, 1)
		
		# gate_proj
		h1 = torch.bmm(W_gate, x_un).squeeze(-1)     # (M, hidden)
		# up_proj
		h2 = torch.bmm(W_up,   x_un).squeeze(-1)     # (M, hidden)

		# elementwise gated activation
		h = self.experts[expert_ids[0]].act_fn(h1) * h2 # (M, hidden)

		# down_proj
		h_un = h.unsqueeze(-1)                       # (M, hidden, 1)
		out = torch.bmm(W_down, h_un).squeeze(-1)    # (M, out)

		# routing weights
		rw = routing_weights[token_pos, local_rank].unsqueeze(-1).to(out.dtype) #out of index here		
		out = out * rw

		# scatter back
		#print("4. expert_hit:", expert_hit.shape, "x:", x.shape, x_un.shape, "W_gate:", W_gate.shape, "W_up:", W_up.shape, out, out.shape)
		final_hidden_states.index_add_(0, token_pos, out.to(final_hidden_states.dtype))
		#=========================================================================

		shared_expert_output = self.shared_expert(hidden_states)
		shared_expert_output = F.sigmoid(self.shared_expert_gate(hidden_states)) * shared_expert_output
		final_hidden_states = final_hidden_states + shared_expert_output
		final_hidden_states = final_hidden_states.reshape(batch_size, sequence_length, hidden_dim)
		return final_hidden_states, router_logits


class MyQwen3NextDecoderLayer(Qwen3NextDecoderLayer, loaderLayer):
	def __init__(self, config, layer_idx):
		super().__init__(config, layer_idx)
		self.layer_idx = layer_idx
		self.mlp.layer_idx = layer_idx
		if hasattr(self.mlp, "experts"):
			for expert_idx, expert_layer in enumerate(self.mlp.experts):
				expert_layer.expert_idx = expert_idx
				expert_layer.layer_idx = layer_idx
				expert_layer._unload_expert_weights()

	def forward(self, *args, **kwargs):
		self._load_layer_weights()
		out = super().forward(*args, **kwargs)
		self._unload_layer_weights()
		return out
	

class MyQwen3NextModel(Qwen3NextModel):
	def __init__(self, config: Qwen3NextConfig):
		super().__init__(config)
		self.config = config
		self.layers = nn.ModuleList()
		for layer_idx in range(config.num_hidden_layers):
			self.layers.append(MyQwen3NextDecoderLayer(config, layer_idx))
			self.layers[-1]._unload_layer_weights()
				
	def forward(
		self,
		input_ids: Optional[torch.LongTensor] = None,
		attention_mask: Optional[torch.Tensor] = None,
		position_ids: Optional[torch.LongTensor] = None,
		past_key_values: Optional[Cache] = None,
		inputs_embeds: Optional[torch.FloatTensor] = None,
		use_cache: Optional[bool] = None,
		cache_position: Optional[torch.LongTensor] = None,
		**kwargs: Unpack[TransformersKwargs],
	) -> MoeModelOutputWithPast:
		if (input_ids is None) ^ (inputs_embeds is not None):
			raise ValueError("You must specify exactly one of input_ids or inputs_embeds")

		if inputs_embeds is None:
			inputs_embeds = self.embed_tokens(input_ids)

		if use_cache and past_key_values is None:
			past_key_values = Qwen3NextDynamicCache(config=self.config)

		if cache_position is None:
			past_seen_tokens = past_key_values.get_seq_length() if past_key_values is not None else 0
			cache_position = torch.arange(
				past_seen_tokens, past_seen_tokens + inputs_embeds.shape[1], device=inputs_embeds.device
			)
		if position_ids is None:
			position_ids = cache_position.unsqueeze(0)

		causal_mask = create_causal_mask(
			config=self.config,
			input_embeds=inputs_embeds,
			attention_mask=attention_mask,
			cache_position=cache_position,
			past_key_values=past_key_values,
			position_ids=position_ids,
		)
		linear_attn_mask = self._update_linear_attn_mask(attention_mask, cache_position)

		hidden_states = inputs_embeds

		# create position embeddings to be shared across the decoder layers
		position_embeddings = self.rotary_emb(hidden_states, position_ids)

		#===============================================
		self.embed_tokens.cpu(); self.parent_lm_head.cpu()
		for decoder_layer in self.layers:
			#print(decoder_layer.layer_idx, "decoder_layer /", self.config.num_hidden_layers, stats.print_and_clean())
			layer_mask = linear_attn_mask if decoder_layer.layer_type == "linear_attention" else causal_mask
			hidden_states = decoder_layer(
				hidden_states,
				position_embeddings=position_embeddings,
				attention_mask=layer_mask,
				position_ids=position_ids,
				past_key_values=past_key_values,
				use_cache=use_cache,
				cache_position=cache_position,
				**kwargs,
			)
		
		hidden_states = self.norm(hidden_states)
		self.embed_tokens.to(hidden_states.device); self.parent_lm_head.to(hidden_states.device)
		if stats: print("./qwen3_next.forward.", datetime.now().strftime("%H:%M:%S"), stats.print_and_clean() if stats else "")
		#================================================

		return MoeModelOutputWithPast(
			last_hidden_state=hidden_states,
			past_key_values=past_key_values,
		)


import transformers.models.qwen3_next.modeling_qwen3_next as modeling
modeling.Qwen3NextMLP = MyQwen3NextMLP
modeling.Qwen3NextSparseMoeBlock = MyQwen3NextSparseMoeBlock
modeling.Qwen3NextModel = MyQwen3NextModel
#===============================================


class MyQwen3NextForCausalLM(Qwen3NextForCausalLM):
	def __init__(self, config):
		super().__init__(config)
		self.model.parent_lm_head = self.lm_head #link
		self.num_hidden_layers = config.num_hidden_layers

	def generate(self, **args):
		with torch.no_grad():			
			return super().generate(**args)

	def offload_layers_to_cpu(self, layers_num=2):
		self.offload_layers_to_gpu_cpu(cpu_layers_num=layers_num)

	def offload_layers_to_gpu_cpu(self, gpu_layers_num=0, cpu_layers_num=0):
		print("offloading layers to CPU/GPU...")
		layer_idx = 0
		while (gpu_layers_num>0 or cpu_layers_num>0) and layer_idx < self.num_hidden_layers:
			base = f"model.layers.{layer_idx}."
			loader.preload_layer_safetensors(base)
			if gpu_layers_num>0:
				loader.offload_dict_to_gpu_cpu(base, gpu=True)
				gpu_layers_num-=1
			else:
				loader.offload_dict_to_gpu_cpu(base, gpu=False)
				cpu_layers_num-=1
			layer_idx+=1
		#import gc; gc.collect()
		print("finished offloading layers to CPU/GPU")


#=============================================================================================
class Qwen3NextMultiTokenPredictor(nn.Module):
	def __init__(self, config):
		super().__init__()
		self.fc = nn.Linear(config.hidden_size*2, config.hidden_size, bias=False)
		self.norm = Qwen3NextRMSNorm(config.hidden_size, eps=config.rms_norm_eps)
		self.pre_fc_norm_embedding = Qwen3NextRMSNorm(config.hidden_size, eps=config.rms_norm_eps)
		self.pre_fc_norm_hidden = Qwen3NextRMSNorm(config.hidden_size, eps=config.rms_norm_eps)
		self.layers = nn.ModuleList([Qwen3NextDecoderLayer(config, layer_idx) for layer_idx in range(1)])

	def forward(self, x, inputs_embeds, position_ids, position_embeddings):
		inputs_embeds = self.pre_fc_norm_embedding(inputs_embeds)
		hidden_states = self.pre_fc_norm_hidden(x)
		hidden_states = torch.cat([inputs_embeds, hidden_states], dim=-1)
		hidden_states = self.fc(hidden_states)
		
		hidden_states = self.layers[0](
			hidden_states,
			position_embeddings=position_embeddings,
			position_ids=position_ids
		)
		hidden_states = self.norm(hidden_states)		

		return MoeModelOutputWithPast(
			last_hidden_state=hidden_states,
			past_key_values=None #past_key_values,
		)


class MyQwen3NextForCausalLM_MTP(MyQwen3NextForCausalLM):
	def __init__(self, config):
		super().__init__(config)
		self.mtp = Qwen3NextMultiTokenPredictor(config)
		self.logits2 = None
		self.past_key_values = None
		self.input_ids, self.cache_position = None, None
	
	def forward(self, **args):
		if self.logits2 is None:
			if self.input_ids: 
				#args["input_ids"] = torch.cat([args["input_ids"], self.input_ids], dim=-1)
				#args["cache_position"] = torch.cat([self.cache_position, args["cache_position"]], dim=-1)
				pass

			outputs = self.model(**args)
			hidden_states, past_key_values = outputs.last_hidden_state, outputs.past_key_values
			logits = self.lm_head(hidden_states[:, -1:, :])
			self.past_key_values = past_key_values

			#MTP
			position_ids = args["cache_position"].unsqueeze(0)			
			position_embeddings = self.model.rotary_emb(hidden_states, position_ids)
			inputs_embeds = self.model.embed_tokens(args["input_ids"]) #double computation, can be taken from main model
			outputs = self.mtp(hidden_states, inputs_embeds, position_ids, position_embeddings)
			hidden_states = outputs.last_hidden_state
			self.logits2 = self.lm_head(hidden_states[:, -1:, :])
			#print(past_key_values, past_key_values.key_cache[0].shape if past_key_values.key_cache[0] else None, "- k shape. hs:", hidden_states.shape)
		else:
			self.input_ids, self.cache_position = args["input_ids"], args["cache_position"]
			logits, past_key_values = self.logits2, self.past_key_values
			self.logits2 = None
		
		return MoeCausalLMOutputWithPast(
			logits=logits,
			past_key_values=past_key_values
		)